//! Koreader Progress Sync API.

use poem::{
    error::ResponseError,
    http::StatusCode,
    web::{Data, Json},
    Error, Result,
};
use poem_openapi::{
    param::Path,
    payload::{self, Json as JsonPayload},
    Object, OpenApi,
};
use serde::{Deserialize, Serialize};
use std::sync::Arc;

use crate::{
    app_state::AppState,
    auth_middleware::{authorize, User},
    db::DataStoreError,
    error_response,
};

/// Data for registering a new user.
#[derive(Object, Deserialize)]
struct RegisterRequest {
    username: String,
    password: String,
}

/// Answer after registering a new user.
#[derive(Object, Serialize)]
struct UserCreated {
    username: String,
}

/// Data for pushing progress.
#[derive(Debug, Clone, Object, Deserialize)]
pub struct DocumentUpdate {
    pub device: String,
    pub device_id: String,
    pub document: String,
    pub percentage: f32,
    pub progress: String,
}

/// Answer when pulling progress.
#[derive(Debug, Clone, Object, Deserialize)]
pub struct DocumentProgress {
    pub device: String,
    pub device_id: String,
    pub percentage: f32,
    pub progress: String,
    pub timestamp: i64,
}

impl ResponseError for DataStoreError {
    fn status(&self) -> StatusCode {
        StatusCode::INTERNAL_SERVER_ERROR
    }

    fn as_response(&self) -> poem::Response
    where
        Self: std::error::Error + Send + Sync + 'static,
    {
        error_response::create_and_log(self, || match self {
            DataStoreError::DatabaseError(_) => (
                "An internal server error ocurred".to_string(),
                StatusCode::INTERNAL_SERVER_ERROR,
            ),
        })
    }
}

pub struct Api;

#[OpenApi]
impl Api {
    /// Register a new user.
    ///
    /// If registration is disabled return early.
    ///
    /// If a user of that id already exist, return a conflict.
    #[oai(path = "/users/create", method = "post")]
    async fn register(
        &self,
        req: Json<RegisterRequest>,
        state: Data<&Arc<AppState>>,
    ) -> Result<payload::Response<payload::Json<UserCreated>>> {
        if !state.config.allow_registration {
            return Err(Error::from_string(
                "Registration is disabled.",
                StatusCode::UNAUTHORIZED,
            ));
        }

        let db = &state.db;
        if db.get_user(&req.username).await?.is_some() {
            Err(Error::from_status(StatusCode::CONFLICT))
        } else {
            db.add_user(&req.username, &req.password).await?;
            Ok(payload::Response::new(payload::Json(UserCreated {
                username: req.username.clone(),
            }))
            .status(StatusCode::CREATED))
        }
    }

    /// Return OK if a user authenticated successfully.
    #[oai(path = "/users/auth", method = "get", transform = "authorize")]
    async fn login(&self) -> Result<payload::Response<()>> {
        Ok(payload::Response::new(()).status(StatusCode::OK))
    }

    /// Push new progress.
    #[oai(path = "/syncs/progress", method = "put", transform = "authorize")]
    async fn push_progress(
        &self,
        state: Data<&Arc<AppState>>,
        user: Data<&User>,
        doc: JsonPayload<DocumentUpdate>,
    ) -> Result<payload::Response<()>> {
        let db = &state.db;

        let doc_update = (user.name.clone(), doc.0).into();
        db.update_position(&doc_update).await?;

        Ok(payload::Response::new(()).status(StatusCode::OK))
    }

    /// Pull progress for document with id `doc_id`.
    #[oai(
        path = "/syncs/progress/:doc_id",
        method = "get",
        transform = "authorize"
    )]
    async fn pull_progress(
        &self,
        state: Data<&Arc<AppState>>,
        user: Data<&User>,
        doc_id: Path<String>,
    ) -> Result<payload::Response<JsonPayload<DocumentProgress>>> {
        let db = &state.db;

        if let Some(doc) = db.get_position(&user.name, &doc_id).await? {
            Ok(payload::Response::new(JsonPayload(DocumentProgress {
                device: doc.device,
                device_id: doc.device_id,
                percentage: doc.percentage,
                progress: doc.progress,
                timestamp: doc.timestamp.assume_utc().unix_timestamp(),
            }))
            .status(StatusCode::OK))
        } else {
            Err(Error::from_status(StatusCode::NOT_FOUND))
        }
    }
}
