//! Poem middleware for koreader progress sync authentication.
//!
//! Authentication works with two headers:
//! - x-auth-user: username
//! - x-auth-key: md5 hashed password

use poem::{
    error::ResponseError, http::StatusCode, Endpoint, EndpointExt, Error, Middleware, Request,
    Result,
};
use std::sync::Arc;
use thiserror::Error as ThisError;

use crate::{app_state::AppState, error_response};

pub fn authorize(ep: impl Endpoint) -> impl Endpoint {
    ep.with(AuthMiddleware {})
}

#[derive(ThisError, Debug)]
pub enum AuthError {
    #[error("failed to extract request data")]
    DataExtractError,
}

#[derive(Debug, Clone)]
pub struct User {
    pub name: String,
}

impl ResponseError for AuthError {
    fn status(&self) -> StatusCode {
        StatusCode::INTERNAL_SERVER_ERROR
    }

    fn as_response(&self) -> poem::Response
    where
        Self: std::error::Error + Send + Sync + 'static,
    {
        error_response::create_and_log(self, || match self {
            AuthError::DataExtractError => (
                "An internal server error ocurred".to_string(),
                StatusCode::INTERNAL_SERVER_ERROR,
            ),
        })
    }
}

struct AuthMiddleware;

impl<E: Endpoint> Middleware<E> for AuthMiddleware {
    type Output = AuthMiddlewareImpl<E>;

    fn transform(&self, ep: E) -> Self::Output {
        AuthMiddlewareImpl { ep }
    }
}

/// The new endpoint type generated by the TokenMiddleware.
struct AuthMiddlewareImpl<E> {
    ep: E,
}

const HEADER_USER: &str = "x-auth-user";
const HEADER_KEY: &str = "x-auth-key";

fn get_header(
    header: &str,
    req: &Request,
    error_msg: &str,
    error_status: StatusCode,
) -> Result<String> {
    Ok(req
        .header(header)
        .ok_or(Error::from_string(error_msg, error_status))?
        .to_string())
}

impl<E: Endpoint> Endpoint for AuthMiddlewareImpl<E> {
    type Output = E::Output;

    async fn call(&self, mut req: Request) -> Result<Self::Output> {
        let username = get_header(
            HEADER_USER,
            &req,
            "No user specified",
            StatusCode::UNAUTHORIZED,
        )?;

        let key = get_header(
            HEADER_KEY,
            &req,
            "No key specified",
            StatusCode::UNAUTHORIZED,
        )?;

        let state = req
            .data::<Arc<AppState>>()
            .ok_or(AuthError::DataExtractError)?;

        let db = &state.db;
        if let Some(user) = db.get_user(&username).await? {
            if user.key == key {
                req.extensions_mut().insert(User { name: username });
                return self.ep.call(req).await;
            }
        }

        Err(Error::from_string(
            "Unauthorized user",
            StatusCode::FORBIDDEN,
        ))
    }
}
