//! Data access and persistence.

use ::entity::{document, user};
use migration::{Migrator, MigratorTrait};
use sea_orm::{
    ActiveModelTrait, ColumnTrait, Database, DatabaseConnection, DbErr, EntityTrait, QueryFilter,
    Set,
};
use thiserror::Error;
use time::{OffsetDateTime, PrimitiveDateTime};
use tracing::debug;

use crate::api::DocumentUpdate;

#[derive(Error, Debug)]
pub enum DataStoreError {
    #[error("database error")]
    DatabaseError(#[from] DbErr),
}

/// Document progress to insert/update.
#[derive(Debug)]
pub struct DocumentInsert {
    id: String,
    user: String,
    device: String,
    device_id: String,
    percentage: f32,
    progress: String,
}

impl From<(String, DocumentUpdate)> for DocumentInsert {
    fn from(value: (String, DocumentUpdate)) -> Self {
        Self {
            id: value.1.document,
            user: value.0,
            device: value.1.device,
            device_id: value.1.device_id,
            percentage: value.1.percentage,
            progress: value.1.progress,
        }
    }
}

pub struct Db {
    connection: DatabaseConnection,
}

impl Db {
    /// Add a new user and return an error if one with `id` already exists.
    pub async fn add_user(&self, id: &str, key: &str) -> Result<user::Model, DataStoreError> {
        let user = user::ActiveModel {
            id: Set(id.to_owned()),
            key: Set(key.to_owned()),
        };
        debug!("saving user {user:?}");
        Ok(user.insert(&self.connection).await?)
    }

    /// Get user with id `id` or None if they do not exist.
    pub async fn get_user(&self, id: &str) -> Result<Option<user::Model>, DataStoreError> {
        Ok(user::Entity::find_by_id(id).one(&self.connection).await?)
    }

    /// Get the progress for document with id `doc_id` from user with id `user_id` or None if that
    /// combination does not exist.
    pub async fn get_position(
        &self,
        user_id: &str,
        doc_id: &str,
    ) -> Result<Option<document::Model>, DataStoreError> {
        Ok(document::Entity::find()
            .filter(document::Column::Id.eq(doc_id))
            .filter(document::Column::User.eq(user_id))
            .one(&self.connection)
            .await?)
    }

    /// Update progress for document with id `doc_id` from user with id `user_id`.
    ///
    /// Set the progress timestamp to now.
    pub async fn update_position(&self, doc: &DocumentInsert) -> Result<(), DataStoreError> {
        let now = OffsetDateTime::now_utc();
        let now = PrimitiveDateTime::new(now.date(), now.time());

        let old_doc = self.get_position(&doc.user, &doc.id).await?;
        let new_doc = document::ActiveModel {
            id: Set(doc.id.clone()),
            user: Set(doc.user.clone()),
            device: Set(doc.device.clone()),
            device_id: Set(doc.device_id.clone()),
            percentage: Set(doc.percentage),
            progress: Set(doc.progress.clone()),
            timestamp: Set(now),
        };

        match old_doc {
            Some(_) => new_doc.update(&self.connection).await?,
            None => new_doc.insert(&self.connection).await?,
        };

        Ok(())
    }
}

/// Connect to the database and return the opened connectoin pool.
pub async fn connect(connection_string: &str) -> Result<Db, DataStoreError> {
    let connection: DatabaseConnection = Database::connect(connection_string).await?;
    Migrator::up(&connection, None).await?;

    Ok(Db { connection })
}
