//! Error handling for the sync service.

use poem::{http::StatusCode, Body, Response};
use poem_openapi::Object;
use serde::Serialize;
use std::error::Error;
use tracing::error;
use uuid::Uuid;

#[derive(Object, Serialize)]
struct ErrorResponse {
    id: String,
    message: String,
}

/// Log `error` on the server side with a generated id and send back a sanitized error message with
/// the same id.
pub fn create_and_log<F>(error: impl Error, cb: F) -> Response
where
    F: Fn() -> (String, StatusCode),
{
    let id = Uuid::new_v4().to_string();
    error!("{id}: {error:?}");

    let (message, status) = cb();
    let response_json = ErrorResponse {
        id: id.clone(),
        message: message.clone(),
    };
    let response_body = Body::from_json(response_json).unwrap_or(Body::from_string(
        format!(
            "{{\"id\": \"{id}\", \"message\": \"{}\" }}",
            message.clone()
        )
        .to_owned(),
    ));

    Response::builder()
        .status(status)
        .header("Content-Type", "application/json")
        .body(response_body)
}
